/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hli@hendriklipka.de
*/

package de.hendriklipka.pdfconv;

import de.hendriklipka.pdfconv.config.ImageConfig;
import de.hendriklipka.pdfconv.config.OutputConfig;
import de.hendriklipka.pdfconv.document.*;
import de.hendriklipka.pdfconv.image.*;
import org.jpedal.PdfDecoder;
import org.jpedal.io.ObjectStore;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Vector;

public class Converter
{
    private OutputConfig _outputConfig=new OutputConfig();
    private ImageConfig _imageConfig=new ImageConfig();

    private Vector _documentWriter=new Vector();
    private Vector _imageWriter=new Vector();
    private ImageProcessor _imageProcessor=null;

    private Vector _pdfFiles=new Vector();

    static DecimalFormat _format=(DecimalFormat)DecimalFormat.getInstance();
    private boolean _debug=false;
    private boolean _verbose=false;
    private Vector _pageList=null;
    private float _pdfScale=1.0f;

    static
    {
        _format.setMinimumIntegerDigits(3);
    }

    public void convert() throws IOException
    {
        //create a directory if it doesn't exist
        File targetPath=new File(_outputConfig.targetDir);
        if (!targetPath.exists())
            targetPath.mkdirs();

        _imageProcessor=new ImageProcessor(_imageConfig);
        _imageProcessor.setVerbose(_verbose);
        _imageProcessor.setDebug(_debug);

        if (null==_outputConfig.baseName)
        {
            for (int i=0; i<_pdfFiles.size(); i++)
            {
                File file=(File)_pdfFiles.elementAt(i);
                String baseName=calculateBaseName(file);
                startConvert(baseName);
                doConvert(file, baseName, 0);
                endConvert(baseName);
            }
        }
        else
        {
            startConvert(_outputConfig.baseName);
            int offset=0;
            for (int i=0; i<_pdfFiles.size(); i++)
            {
                File file=(File)_pdfFiles.elementAt(i);
                int pages=doConvert(file, _outputConfig.baseName, offset);
                offset+=pages;
            }

            endConvert(_outputConfig.baseName);
        }
    }

    private String calculateBaseName(File file)
    {
        String s=file.getName();
        int pos=s.lastIndexOf('.');
        if (-1==pos)
            return s;
        return s.substring(0, pos);
    }

    private void startConvert(String baseName) throws IOException
    {
        for (int i=0; i<_documentWriter.size(); i++)
        {
            DocumentWriter writer=(DocumentWriter)_documentWriter.elementAt(i);
            writer.open(_outputConfig.targetDir+baseName);
            writer.writeHeader(baseName);
        }
    }

    private void endConvert(String baseName) throws IOException
    {
        for (int i=0; i<_documentWriter.size(); i++)
        {
            DocumentWriter writer=(DocumentWriter)_documentWriter.elementAt(i);
            writer.writeFooter();
            writer.close();
        }
    }

    private int doConvert(File pdfFile, String baseName, int startNum) throws IOException
    {
        long startTime=System.currentTimeMillis();
        long docStartTime=System.currentTimeMillis();
        
        File f=new File(_outputConfig.targetDir);
        _outputConfig.targetDir=f.getCanonicalPath()+File.separator;
        int pageCount=0;

        PdfDecoder pdfDecoder=null;
        String pdfName=pdfFile.getCanonicalPath();

        if (_verbose)
            System.out.println("converting PDF file: "+pdfName);
        try
        {
            ObjectStore.init("pdfconv");
            pdfDecoder=new PdfDecoder(true);
            pdfDecoder.openPdfFile(pdfName);
            pdfDecoder.setScreenRes(_pdfScale);
        }
        catch (Exception ex)
        {
            System.err.println("Error while opening PDF file");
            ex.printStackTrace();
            System.exit(1);
        }

        try
        {
            pageCount=pdfDecoder.getPageCount();
            if (_verbose)
                System.out.println("found "+pageCount+" pages");
            long currTime=System.currentTimeMillis();
            if (_verbose)
                System.out.println("time to open file: "+(currTime-startTime)/1000+" seconds");
            startTime=System.currentTimeMillis();
            for (int page=1; page<=pageCount; page++)
            {
                if (null!=_pageList)
                {
                    boolean found=false;
                    for (int i=0; i<_pageList.size(); i++)
                    {
                        Object o=_pageList.elementAt(i);
                        if (o instanceof String)
                        {
                            String s=(String)o;
                            if (s.endsWith("-"))
                            {
                                s=s.substring(0,s.length()-1);
                                try
                                {
                                    int pageNum=Integer.parseInt(s);
                                    if (pageNum<=page)
                                            found=true;
                                }
                                catch (NumberFormatException e)
                                {
                                }
                            }
                        }
                        else
                        {
                            Integer pageNum=(Integer)_pageList.elementAt(i);
                            if (page==pageNum.intValue())
                                found=true;
                        }
                    }
                    if (!found)
                    {
                        if (_verbose)
                        {
                            System.out.println();
                            System.out.println("skipping page #"+page);
                        }
                        continue;
                    }
                }
                long pageStartTime=System.currentTimeMillis();
                String pageNum=_format.format(page+startNum);
                if (_verbose)
                {
                    System.out.println();
                    System.out.println("Processing page #"+page+"/"+pageCount);
                }

                System.gc();
                
                String targetName=baseName+"_page_"+pageNum;
                pdfDecoder.initDisplay(page);
                int width=pdfDecoder.getMaxX(page)-pdfDecoder.getMinX(page);
                int heigth=pdfDecoder.getMaxY(page)-pdfDecoder.getMinY(page);
                if (0!=_imageWriter.size())
                {
                    try
                    {
                        pdfDecoder.decodePage(true, page);

                        BufferedImage targetImage=pdfDecoder.getPageAsImage();
                        currTime=System.currentTimeMillis();
                        if (_verbose)
                            System.out.println("time to render page: "+(currTime-startTime)/1000+" seconds");
                        startTime=System.currentTimeMillis();
                        if (null==targetImage)
                        {
                            System.err.println("can't render page, skipping");
                        }
                        else
                        {
                            if (_verbose)
                                System.out.println("post-processing image");
                            targetImage=_imageProcessor.processImage(targetImage);
                            currTime=System.currentTimeMillis();
                            if (_verbose)
                                System.out.println("time to process image: "+(currTime-startTime)+" ms");
                            startTime=System.currentTimeMillis();
                            for (int i=0; i<_imageWriter.size(); i++)
                            {
                                ImageWriter imageWriter=(ImageWriter)_imageWriter.elementAt(i);
                                if (_verbose)
                                    System.out.println("saving image as "+imageWriter.getImageType());
                                imageWriter.saveImage(_outputConfig.targetDir+targetName, targetImage, _imageConfig);
                                currTime=System.currentTimeMillis();
                                if (_verbose)
                                    System.out.println("time to write image: "+(currTime-startTime)+" ms");
                                startTime=System.currentTimeMillis();
                            }

                        }
                    }
                    catch (Exception ex)
                    {
                        System.err.println("Error while writing image");
                        ex.printStackTrace();
                    }
                    finally
                    {
                        pdfDecoder.clearPdfImageData();
                        System.gc();
                    }
                }
                for (int i=0; i<_documentWriter.size(); i++)
                {
                    DocumentWriter writer=(DocumentWriter)_documentWriter.elementAt(i);
                    writer.addImage(targetName, width, heigth);
                }
                currTime=System.currentTimeMillis();
                if (_verbose)
                    System.out.println("page time: "+(currTime-pageStartTime)/1000+" seconds");
            }
        }
        catch (Exception ex)
        {
            pdfDecoder.closePdfFile();
            System.err.println("Error while executing ");
            ex.printStackTrace();
            System.exit(1);
        }
        finally
        {
            if (null!=pdfDecoder)
            {
                pdfDecoder.markAllPagesAsUnread();
                pdfDecoder.flushObjectValues(false);
                pdfDecoder.closePdfFile();
            }
            ObjectStore.flush();
        }
        long currTime=System.currentTimeMillis();
        if (_verbose)
        {
            System.out.println();
            System.out.println("document time: "+(currTime-docStartTime)/1000+" seconds");
        }
        return pageCount;
    }

    public void setWriteBMP()
    {
        _imageWriter.addElement(new BMPImageWriterGray());
    }

    public void setWriteBMP2()
    {
        _imageWriter.addElement(new BMPImageWriter());
    }

    public void setWritePICT()
    {
        _imageWriter.addElement(new PICTImageWriter());
    }

    public void setWriteJPEG()
    {
        _imageWriter.addElement(new JPEGImageWriter());
    }

    public void setWriteHTML()
    {
        _documentWriter.addElement(new HTMLWriter());
    }

    public void setWriteBookMac()
    {
        _documentWriter.addElement(new BookWriterMac());
    }

    public void setWriteBookWin()
    {
        _documentWriter.addElement(new BookWriterWin());
    }

    public void setQuality(int value)
    {
        _imageConfig.quality=value;
    }

    public void setScaleFactor(double value)
    {
        _imageConfig.scaleFactor=value;
    }

    public void setMaxHeigth(int value)
    {
        _imageConfig.maxHeigth=value;
    }

    public void setMaxWidth(int value)
    {
        _imageConfig.maxWidth=value;
    }

    public void setTargetDir(String targetDir)
    {
        _outputConfig.targetDir=targetDir;
    }

    public void setBaseName(String targetBaseName)
    {
        _outputConfig.baseName=targetBaseName;
    }

    public void addSourceFile(File file)
    {
        _pdfFiles.addElement(file);
    }

    public void setDebug(boolean set)
    {
        _debug=set;
    }

    public void setVerbose(boolean set)
    {
        _verbose=set;
    }

    public void setWriteNTKMac()
    {
        _documentWriter.addElement(new NTKWriterMac());
    }

    public void setWriteNTKWin()
    {
        _documentWriter.addElement(new NTKWriterWin());
    }
    
    public void setPagesToConvert(Vector pageList)
    {
        _pageList=pageList;
    }

    public void setPDFScale(float pdfScale)
    {
        _pdfScale=pdfScale;
    }
}
